#pragma once
#include <ConsumerImplHelper/Platform.h>
#include <ConsumerImplHelper/BufferAllocator.h>
#include <ConsumerImplHelper/ProducerProxy.h>

#include <memory>

namespace GenTLConsumerImplHelper
{
    template< const char ProducerName[], typename ProducerProxy>
    class CGenTLCamera;

    namespace Internal
    {
        // Does the clean-up and frees the resources if image acquisition is stopped or 
        // interrupted.
        template< const char ProducerName[], typename ProducerProxy = CProducerProxy<ProducerName>>
        class CGrabResources
        {
            using Camera_t = CGenTLCamera<ProducerName, ProducerProxy>;

        public:
            class Buffer
            {
            public:
                Buffer()
                    : m_pBuffer(nullptr)
                    , m_handle(GENTL_INVALID_HANDLE)
                    , m_isQueued(false)
                {
                }

                void* getBuffer() const { return m_pBuffer; }
                void setBuffer(void *pBuffer) { m_pBuffer = pBuffer; }

                GenTL::BUFFER_HANDLE& getHandle() { return m_handle;  }
                const GenTL::BUFFER_HANDLE& getHandle() const { return m_handle; }

                void setQueued(bool isQueued) { m_isQueued = isQueued;  }
                bool isQueued() const { return m_isQueued;  }

            private:
                void*                   m_pBuffer;
                GenTL::BUFFER_HANDLE    m_handle;
                bool                    m_isQueued;
            };

        public:
            enum cleanupLevel
            {
                DoNothing,
                FreeBuffers,
                RevokeBuffers,
                UnregisterGrabEvent,
                FlushQueues,
                StopProducer,
                StopCamera
            };

            CGrabResources(size_t nBuffers, Camera_t& parent, GenTLConsumerImplHelper::BufferAllocator& allocator)
                : m_level(DoNothing)
                , m_isDestructing(false)
                , m_buffers(nBuffers)
                , m_parent(parent)
                , m_pAllocator(&allocator)
                , m_hNewBufferEvent(GENTL_INVALID_HANDLE)

            {
            }

            CGrabResources(Camera_t& parent, std::vector<Buffer>& buffers)
                : m_level(DoNothing)
                , m_isDestructing(false)
                , m_buffers(buffers)
                , m_parent(parent)
                , m_pAllocator(nullptr)
                , m_hNewBufferEvent(GENTL_INVALID_HANDLE)
            {
            }


            ~CGrabResources()
            {
                m_isDestructing = true;  // from now on, we are going to ignore setCleanupLevel calls issued while shutting down everything.
                using namespace GenTL;

                GenTLConsumerImplHelper::CProducerProxy<ProducerName> &producer = m_parent.GetProducer();
                DS_HANDLE hDataStream = m_parent.GetDataStreamHandle();

                if (m_level >= StopCamera)
                {
                    try
                    {
                        m_parent.IssueAcquisitionStopCommand();
                    }
                    catch (const GenICam::GenericException&)
                    {
                    }
                }

                if (m_level >= StopProducer)
                {
                    try
                    {
                        m_parent.StopAcquisition();
                    }
                    catch (const GenICam::GenericException&)
                    {
                    }
                }

                if (m_level >= UnregisterGrabEvent && m_hNewBufferEvent != GENTL_INVALID_HANDLE)
                {
                    producer.GCUnregisterEvent(hDataStream, EVENT_NEW_BUFFER);
                    m_hNewBufferEvent = GENTL_INVALID_HANDLE;
                }

                if (m_level >= FlushQueues)
                {
                    producer.DSFlushQueue(hDataStream, ACQ_QUEUE_ALL_DISCARD);
                }
                if (m_level >= RevokeBuffers)
                {
                    for (size_t i = 0; i < m_buffers.size(); ++i)
                    {
                        GenTL::BUFFER_HANDLE& handle(m_buffers[i].getHandle());
                        if (handle  != GENTL_INVALID_HANDLE)
                        {
                            void* pPrivate;
                            void* pBuffer;
                            producer.DSRevokeBuffer(hDataStream, handle, &pBuffer, &pPrivate);
                            handle = GENTL_INVALID_HANDLE;
                            m_buffers[i].setQueued(false);
                        }
                    }
                }
                if (m_level >= FreeBuffers && m_pAllocator != nullptr)
                {
                    for (size_t i = 0; i < m_buffers.size(); ++i)
                    {
                        void* pBuffer = m_buffers[i].getBuffer();
                        if (pBuffer != nullptr )
                        {
                            m_buffers[i].setBuffer(nullptr);
                            m_pAllocator->FreeBuffer(pBuffer);
                        }
                    }
                }
                m_level = DoNothing;
            }

            std::vector<Buffer>&                GetBufferPtrs() { return m_buffers; }
            Buffer&                             GetBuffer(GenTL::BUFFER_HANDLE h) { return m_buffers[LookupBufferHandle(h)]; }
            GenTL::EVENT_HANDLE&                GetNewBufferEvent() { return m_hNewBufferEvent; }

            void                                SetCleanupLevel(cleanupLevel level) { if ( ! m_isDestructing ) { m_level = level; } }
            cleanupLevel                        GetCleanupLevel() const { return m_level; }

        protected:
            size_t LookupBufferHandle(GenTL::BUFFER_HANDLE h)
            {
                size_t idx = 0;
                for (idx = 0; idx < m_buffers.size(); ++idx)
                {
                    if (m_buffers[idx].getHandle() == h)
                    {
                        return idx;
                    }
                }
                throw LOGICAL_ERROR_EXCEPTION("Invalid handle passed to QueueBuffer");
            }

        private:
            CGrabResources(const CGrabResources& );
            CGrabResources& operator=(const CGrabResources& );

        private:
            cleanupLevel                                                m_level;
            bool                                                        m_isDestructing;
            std::vector<Buffer>                                         m_buffers;
            Camera_t&                                                   m_parent;
            GenTLConsumerImplHelper::BufferAllocator*                   m_pAllocator;
            GenTL::EVENT_HANDLE                                         m_hNewBufferEvent;

            friend Camera_t;
        };
    }
}
