/*
 * Copyright (c) 2020, Basler AG
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions, and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions, and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of the copyright holder nor the
 *      names of its contributors may be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <ros/ros.h>

#include <basler_blaze/SetBlazeParameter.h>
#include <basler_blaze/GetBlazeParameter.h>
#include <basler_blaze/SetBlazeDataStreamParameter.h>
#include <basler_blaze/GetBlazeDataStreamParameter.h>
#include <basler_blaze/ExecuteCommand.h>

#include <sensor_msgs/Image.h>
#include <camera_info_manager/camera_info_manager.h>

#include <cv_bridge/cv_bridge.h>
#include <opencv2/highgui/highgui.hpp>

#include <pcl_ros/point_cloud.h>
#include <pcl/point_types.h>

// Typedefs for the PCL types used
typedef pcl::PointXYZRGB Point_t;
typedef pcl::PointCloud<Point_t> PointCloud_t;
typedef PointCloud_t::Ptr PointCloudPtr;

#include <iostream>
#include <iomanip>
#include <atomic>
#include <functional>
using namespace std;

#include "ConsumerImplHelper/BlazeCamera.h"
using namespace GenTLConsumerImplHelper;
using namespace GenApi;

#pragma pack(push, 1)
struct BGR 
{
    uint8_t b;
    uint8_t g;
    uint8_t r;
};
#pragma pack(pop)

// blaze camera
class BlazeCamera
{
public:
  BlazeCamera() : m_cameraId("") {}

  bool SetParameter(basler_blaze::SetBlazeParameter::Request  &req,
                    basler_blaze::SetBlazeParameter::Response &res);
  bool GetParameter(basler_blaze::GetBlazeParameter::Request  &req,
                    basler_blaze::GetBlazeParameter::Response &res);
  bool SetDataStreamParameter(basler_blaze::SetBlazeDataStreamParameter::Request  &req,
                              basler_blaze::SetBlazeDataStreamParameter::Response &res);
  bool GetDataStreamParameter(basler_blaze::GetBlazeDataStreamParameter::Request  &req,
                              basler_blaze::GetBlazeDataStreamParameter::Response &res);
  bool ExecuteCommand(basler_blaze::ExecuteCommand::Request  &req,
                      basler_blaze::ExecuteCommand::Response &res);

  sensor_msgs::CameraInfoPtr GetCameraInfo();

  int InitializeCamera(std::string device_id);
  void TerminateProcess();

  int Run();

private:
  template <typename T>
  void roundToIncrement(T &param, T min, T inc);

  bool getAndCheckParameter(const char* name,
                            std::function<GenApi::INode*(const char*)> paramGetter,
                            GenApi::CValuePtr& ptrParameter, std::string& statusMsg);

  template<class T, class Ptr>
  bool setNumericalValue(GenApi::CValuePtr ptrGeneric, T& value, std::string& statusMsg);

  template<typename RequestT, typename ResponseT>
  bool setValue(std::function<GenApi::INode*(const char*)> paramGetter,
      const RequestT& req, ResponseT& resp);

  template<typename ResponseT>
  bool getValue(std::function<GenApi::INode*(const char*)> parameterGetter, const std::string& paramID, ResponseT& res);

  // Checks whether a given 3D point represents a valid coordinate.
  // The Scan3dInvalidDataValue is used to identify a non-valid pixel.
  inline bool isValid(const CBlazeCamera::Coord3D* point)
  {
    static constexpr bool isInvalidValueNaN = s_invalidDataValue != s_invalidDataValue; // true, when s_invalidDataValue equals NaN
    return isInvalidValueNaN ? !std::isnan(point->z) : point->z != s_invalidDataValue;
  };
  
  // Calculates a grayscale depth map from point cloud data sent from a blaze camera.
  // The buffer that pRangeMap points to must be allocated accordingly before 
  // passing it to the calculateRangeMap function.
  void calculateRangeMap(const PartInfo& pointCloud, int minDepth, int maxDepth, uint16_t* pRangeMap);
  // Calculates a color depth map from point cloud data sent from a blaze camera.
  // The buffer that pRangeMap points to must be allocated accordingly before
  // passing it to the calculateRangeMap function.
  void calculateRangeMapColor(const PartInfo& pointCloud, int minDepth, int maxDepth, BGR* pRangeMap);
  //Convert the Basler point cloud to a PCL point cloud.
  PointCloudPtr convertGrabResultToPointCloud(const BufferParts& imgParts);

  // Publish results.
  bool publish(const BufferParts& parts, ros::Time acquisition_time);

private:
  CBlazeCamera    m_camera;

  // Value that identifies pixel with missing depth information.
  constexpr static double  s_invalidDataValue = std::numeric_limits<double>::quiet_NaN(); 
  // Remember current setting in order to restore it when node is shut down.
  double                   m_invalidDataValue_old = 0.0f;

public:
  // Camera ID
  std::string m_cameraId;

  // Topics
  ros::Publisher m_cloudPub;
  ros::Publisher m_intensityPub;
  ros::Publisher m_rangeMapPub;
  ros::Publisher m_rangeMapColorPub;
  ros::Publisher m_confidencePub;

  ros::Publisher m_cloudCIPub;
  ros::Publisher m_intensityCIPub;
  ros::Publisher m_rangeMapCIPub;
  ros::Publisher m_rangeMapColorCIPub;
  ros::Publisher m_confidenceCIPub;

  // Camera info manager
  boost::shared_ptr<camera_info_manager::CameraInfoManager> m_cloudInfoManager;
  boost::shared_ptr<camera_info_manager::CameraInfoManager> m_intensityInfoManager;
  boost::shared_ptr<camera_info_manager::CameraInfoManager> m_rangeMapInfoManager;
  boost::shared_ptr<camera_info_manager::CameraInfoManager> m_rangeMapColorInfoManager;
  boost::shared_ptr<camera_info_manager::CameraInfoManager> m_confidenceInfoManager;
};

// Node features
const std::string node_name = "basler_blaze_node";

// Device serial number
std::string g_deviceId;
// Reference frame ID
std::string g_frameId;
// Namespace specified by user
std::string g_namespace;

// Topic names
std::string g_cloudTopicName = "cloud";
std::string g_intensityTopicName = "intensity";
std::string g_rangeMapTopicName = "range_map";
std::string g_rangeMapColorTopicName = "range_map_color";
std::string g_confidenceTopicName = "confidence";

std::string g_cloudCITopicName = g_cloudTopicName + "/camera_info";
std::string g_intensityCITopicName = g_intensityTopicName + "/camera_info";
std::string g_rangeMapCITopicName = g_rangeMapTopicName + "/camera_info";
std::string g_rangeMapColorCITopicName = g_rangeMapColorTopicName + "/camera_info";
std::string g_confidenceCITopicName = g_confidenceTopicName + "/camera_info";

// Service names
std::string g_setParameterSrvName = "set_blaze_parameter";
std::string g_getParameterSrvName = "get_blaze_parameter";
std::string g_setDSParameterSrvName = "set_blaze_data_stream_parameter";
std::string g_getDSParameterSrvName = "get_blaze_data_stream_parameter";
std::string g_execCmdSrvName = "execute_command";

template <typename T>
void BlazeCamera::roundToIncrement(T &param, T min, T inc)
{
  // Allowed value must follow value ==  min + n * inc, n integer.
  param = min + ((int)((param - min) / inc)) * inc;
}

bool BlazeCamera::getAndCheckParameter(const char* name,
                                       std::function<GenApi::INode*(const char*)> paramGetter,
                                       GenApi::CValuePtr& ptrParameter, std::string& statusMsg)
{
  CNodePtr ptrNode = paramGetter(name);

  // Check validity.
  if (!ptrNode.IsValid())
  {
    ROS_WARN_STREAM("There is no " << name << " parameter.");
    statusMsg = std::string() + "There is no " + name + " parameter.";
    return false;
  }

  if (!GenApi::IsAvailable(ptrNode))
  {
      ROS_WARN_STREAM("The current state of the device doesn't allow access to the " << name << " parameter.");
      statusMsg = std::string() + "The current state of the device doesn't allow access to the " + name + " parameter.";
      return false;
  }

  ptrParameter = ptrNode; // Casts CNodePtr to CValuePtr.
  if (!ptrParameter.IsValid())
  {
    // Parameter is not a value.
    ROS_WARN_STREAM("Parameter " << name << " does not represent a value.");
    statusMsg = std::string() + "Parameter " + name + " does not represent a value.";
    return false;
  }
  return true;
}

template<class T, class Ptr>
bool BlazeCamera::setNumericalValue(GenApi::CValuePtr ptrGeneric, T& value, std::string& statusMsg)
{
  try
  {
    Ptr ptrValue(ptrGeneric);
    if ( ! ptrValue.IsValid() )
    {
        // Shouldn't be possible to reach this point
        // since caller should have already checked principal interface, i.e., the
        // parameter's type.
        statusMsg = "Logical error. Unexpected type of parameter ";
    }

    // Clip value to boundaries and adjust it according to the parameter's increments.
    const T min = static_cast<T>(ptrValue->GetMin());
    const T max = static_cast<T>(ptrValue->GetMax());

    if (value < min)
    {
      ROS_WARN_STREAM("Desired value of " << value << " is smaller than the current minimum value of the "
                       << ptrValue->GetNode()->GetName() << " parameter (" << min << "). Value will be clipped.");
      value = min;
    }
    else if (value > max)
    {
      ROS_WARN_STREAM("Desired value of " << value << " is greater than the current maximum value of the "
                       << ptrValue->GetNode()->GetName() << " parameter (" << max << "). Value will be clipped.");
      value = max;
    }

    const T inc = static_cast<T>(ptrValue->GetInc());
    if (inc != 1)
    {
      roundToIncrement(value, min, inc);
    }

    ptrValue->SetValue(value);
  }
  catch (const GenICam::GenericException& e)
  {
    ROS_ERROR_STREAM("Exception occurred when setting parameter value: " << e.GetDescription());
    statusMsg = std::string() + "Exception occurred when setting parameter value: " + e.GetDescription();
    return false;
  }
  catch (const std::exception& e)
  {
    ROS_ERROR_STREAM("Exception occurred when setting parameter value: " << e.what());
    statusMsg = std::string() + "Exception occurred when setting parameter value: " + e.what();
    return false;
  }
  catch ( ... )
  {
    ROS_ERROR("Unknown exception occurred when setting parameter value.");
    statusMsg = std::string() + "Exception occurred when setting parameter value.";
    return false;
  }

  return true;
}

template<typename RequestT, typename ResponseT>
bool BlazeCamera::setValue(std::function<GenApi::INode*(const char*)> paramGetter,
                           const RequestT& req, ResponseT& res)
{
    try
    {
        std::string paramID = req.ParameterID;
        std::string paramValue = req.ParameterValue;
        ROS_INFO("[%s] Setting parameter %s to %s", ros::this_node::getName().c_str(), paramID.c_str(), paramValue.c_str());

        GenApi::CValuePtr ptrValue;
        std::string statusMsg;

        if ( ! getAndCheckParameter(paramID.c_str(), paramGetter, ptrValue, statusMsg) )
        {
            res.StatusValue = 0;
            res.StatusMsg = statusMsg;
            return true;
        }

        if (!GenApi::IsWritable(ptrValue))
        {
            ROS_WARN_STREAM("The '" << paramID << "' parameter is read-only.");
            statusMsg = std::string("The '") + paramID + "' parameter is read-only.";
            return true;
        }

        const auto type = ptrValue->GetNode()->GetPrincipalInterfaceType();

        if (type == GenApi::intfIInteger)
        {
            int64_t val = std::stoll(paramValue);
            if (!setNumericalValue<int64_t, GenApi::CIntegerPtr>(ptrValue, val, statusMsg))
            {
                res.StatusValue = 0;
                res.StatusMsg = statusMsg;
            }
            else
            {
                res.StatusValue = 1;
                res.StatusMsg = "Set parameter '" + paramID + "' to " + std::to_string(val);
            }
        }
        else if (type == GenApi::intfIFloat)
        {
            double val = stod(paramValue);
            if (!setNumericalValue<double, GenApi::CFloatPtr>(ptrValue, val, statusMsg))
            {
                res.StatusValue = 0;
                res.StatusMsg = statusMsg;
            }
            else
            {
                res.StatusValue = 1;
                res.StatusMsg = "Set parameter '" + paramID + "' to " + std::to_string(val);
            }
        }
        else
        {
            ptrValue->FromString( paramValue.c_str() );
            res.StatusValue = 1;
            res.StatusMsg = "Set parameter '" + paramID + "' to " + paramValue;
        }
    }
    catch(GenICam::GenericException& e)
    {
        ROS_ERROR_STREAM("Exception occurred when setting parameter " << req.ParameterID << ":" << endl << e.GetDescription());
        res.StatusValue = 0;
        res.StatusMsg = std::string("Exception occurred: ") + e.GetDescription();
    }
    catch (const std::exception& e)
    {
        ROS_ERROR_STREAM("Exception occurred when setting parameter " << req.ParameterID << ": " << endl << e.what());
        res.StatusValue = 0;
        res.StatusMsg = std::string("Exception occurred: ") + e.what();
    }
    catch ( ... )
    {
        ROS_ERROR_STREAM("Unknown exception occurred when setting parameter " << req.ParameterID << ".");
        res.StatusValue = 0;
        res.StatusMsg = "Exception occurred.";
    }

    return true;
}

template<typename ResponseT>
bool BlazeCamera::getValue(std::function<GenApi::INode*(const char*)> parameterGetter, const std::string& paramID,
                           ResponseT& res)
{
    try
    {
        GenApi::CValuePtr ptrValue;
        std::string statusMsg;

        if ( ! getAndCheckParameter(paramID.c_str(), parameterGetter, ptrValue, statusMsg) )
        {
            res.StatusValue = 0;
            res.StatusMsg = statusMsg;
            return true;
        }

        if ( ! GenApi::IsReadable(ptrValue))
        {
            res.StatusValue = 0;
            res.StatusMsg = "The parameter '" + paramID + "' is not readable.";
            return true;
        }

        const auto type = ptrValue->GetNode()->GetPrincipalInterfaceType();
        std::string paramValueStr;

        if (type == GenApi::intfIInteger)
        {
            paramValueStr = std::to_string(GenApi::CIntegerPtr(ptrValue)->GetValue());
        }
        else if (type == GenApi::intfIFloat)
        {
            paramValueStr = std::to_string(GenApi::CFloatPtr(ptrValue)->GetValue());
        }
        else
        {
            paramValueStr = ptrValue->ToString().c_str();
        }
        ROS_INFO("[%s] Getting %s parameter value: %s", ros::this_node::getName().c_str(), paramID.c_str(), paramValueStr.c_str());

        res.ParameterValue = paramValueStr;
        res.StatusValue = 1;
        res.StatusMsg = "Value of parameter '" + paramID + "' is " + paramValueStr;
    }
    catch(GenICam::GenericException& e)
    {
        ROS_ERROR_STREAM("Exception occurred when getting parameter '" << paramID << "': "
            << endl << e.GetDescription());
        res.StatusValue = 0;
        res.StatusMsg = std::string("Exception occurred: ") + e.GetDescription();
    }
    catch (const std::exception& e)
    {
        ROS_ERROR_STREAM("Exception occurred when getting parameter '" << paramID << "': "
            << endl << e.what());
        res.StatusValue = 0;
        res.StatusMsg = std::string("Exception occurred: ") + e.what();
    }
    catch (...)
    {
        ROS_ERROR_STREAM("Unknown exception occurred when getting parameter '" << paramID << "'.");
        res.StatusValue = 0;
        res.StatusMsg = std::string("Unknown exception occurred.");
    }
    return true;
}

bool BlazeCamera::GetParameter(basler_blaze::GetBlazeParameter::Request  &req,
                               basler_blaze::GetBlazeParameter::Response &res)
{
    auto paramGetter = [this](const char* name){return m_camera.GetParameter(name);};
    return getValue(paramGetter, req.ParameterID, res);
}

bool BlazeCamera::GetDataStreamParameter(basler_blaze::GetBlazeDataStreamParameter::Request  &req,
                                         basler_blaze::GetBlazeDataStreamParameter::Response &res)
{
    auto paramGetter = [this](const char* name){return m_camera.GetDataStreamParameter(name); };
    return getValue(paramGetter, req.ParameterID, res);
}

bool BlazeCamera::SetParameter(basler_blaze::SetBlazeParameter::Request  &req,
                               basler_blaze::SetBlazeParameter::Response &res) {
    auto getter = [this](const char *name) {
        return m_camera.GetParameter(name);
    };
    return setValue(getter, req, res);
}

bool BlazeCamera::SetDataStreamParameter(basler_blaze::SetBlazeDataStreamParameter::Request  &req,
                                         basler_blaze::SetBlazeDataStreamParameter::Response &res) {
    auto getter = [this](const char *name) {
        return m_camera.GetDataStreamParameter(name);
    };
    return setValue(getter, req, res);
}

bool BlazeCamera::ExecuteCommand(basler_blaze::ExecuteCommand::Request  &req,
                                 basler_blaze::ExecuteCommand::Response &res)
{
  try 
  {
      const std::string cmd = req.Command;
      ROS_INFO("[%s] Executing command %s", ros::this_node::getName().c_str(), cmd.c_str());

      GenApi::CValuePtr ptrValue;
      std::string statusMsg;

      auto paramGetter = [this](const char* name) { return m_camera.GetParameter(name); };
      if ( ! getAndCheckParameter(cmd.c_str(), paramGetter, ptrValue, statusMsg) )
      {
          res.StatusValue = 0;
          res.StatusMsg = statusMsg;
          return true;
      }

      GenApi::CCommandPtr ptrCommand(m_camera.GetParameter(cmd.c_str()));
      if ( ! ptrCommand.IsValid() )
      {
          res.StatusValue = 0;
          res.StatusMsg = cmd + " is not a valid command.";
          return true;
      }

      ptrCommand->Execute();
      res.StatusValue = 1;
      res.StatusMsg = "Executed command " + cmd;
  }
  catch(GenICam::GenericException& e)
  {
    ROS_ERROR_STREAM("Exception occurred when executing command " << req.Command
        << ":" << endl << e.GetDescription());
    res.StatusValue = 0;
    res.StatusMsg = std::string("Exception occurred: ") + e.GetDescription();
  }
  catch (const std::exception& e)
  {
      ROS_ERROR_STREAM("Exception occurred when executing command " << req.Command
      << ":" << endl << e.what());
    res.StatusValue = 0;
    res.StatusMsg = std::string("Exception occurred: ") + e.what();
  }
  catch ( ... )
  {
    ROS_ERROR_STREAM("Unknown exception occurred when executing command " << req.Command);
    res.StatusValue = 0;
    res.StatusMsg = std::string("Exception occurred when executing command.");
  }

  return true;
}

sensor_msgs::CameraInfoPtr BlazeCamera::GetCameraInfo()
{
    sensor_msgs::CameraInfoPtr ci;

    if ( ! m_camera.IsOpen() )
    {
        ROS_ERROR("blaze camera must be opened before retrieving the camera info.");
        return ci;
    }

    try {
        sensor_msgs::CameraInfoPtr newci(new sensor_msgs::CameraInfo);
        newci->width = static_cast<uint32_t>(GenApi::CIntegerPtr(m_camera.GetParameter("Width"))->GetValue());
        newci->height = static_cast<uint32_t>(GenApi::CIntegerPtr(m_camera.GetParameter("Height"))->GetValue());
        const double f = GenApi::CFloatPtr(m_camera.GetParameter("Scan3dFocalLength"))->GetValue();
        const double cx = GenApi::CFloatPtr(m_camera.GetParameter("Scan3dPrincipalPointU"))->GetValue();
        const double cy = GenApi::CFloatPtr(m_camera.GetParameter("Scan3dPrincipalPointV"))->GetValue();

        newci->distortion_model = "plumb_bob";
        newci->D = {0.0, 0.0, 0.0, 0.0, 0.0};
        newci->K = { f, 0.0, cx,
                    0.0, f, cy,
                    0.0, 0.0, 1.0 };
        newci->R = {  1.0, 0.0, 0.0,
                      0.0, 1.0, 0.0,
                      0.0, 0.0, 1.0};
        newci->P = { f, 0.0, cx, 0.0,
                     0.0, f, cy, 0.0,
                     0.0, 0.0, 1.0, 0.0 };

        ci.swap(newci);
    }
    catch ( GenICam::GenericException& e)
    {
        ROS_ERROR_STREAM("Failed to retrieve calibration data from camera: " << e.GetDescription());
    }
    return ci;
}

int BlazeCamera::InitializeCamera(std::string deviceId)
{
  int exitCode = EXIT_SUCCESS;

  try
  {
    CBlazeCamera::InitProducer();

    CameraList cameraList = m_camera.EnumerateCameras();
    bool found = false;

    ROS_INFO("[%s] Found %i connected camera device(s)", ros::this_node::getName().c_str(), (int)cameraList.size());
    int camIndex = 1;
    for (CameraList::const_iterator it = cameraList.begin(); it != cameraList.end(); ++it)
    {
      //ROS_INFO("[%s] * Camera index #%i", ros::this_node::getName().c_str(), camIndex);
      //ROS_INFO("[%s] *** ID:                %s", ros::this_node::getName().c_str(), it->strDeviceID.c_str());
      //ROS_INFO("[%s] *** Model name:        %s", ros::this_node::getName().c_str(), it->strModelName.c_str());
      //ROS_INFO("[%s] *** Display name:      %s", ros::this_node::getName().c_str(), it->strDisplayName.c_str());
      //ROS_INFO("[%s] *** Serial number:     %s", ros::this_node::getName().c_str(), it->strSerialNumber.c_str());
      //ROS_INFO("[%s] *** User-defined name: %s", ros::this_node::getName().c_str(), it->strUserDefinedName.c_str());
      //ROS_INFO("[%s] *** IP address:        %s", ros::this_node::getName().c_str(), it->strIpAddress.c_str());

      // Specified camera has been found.
      if (it->strSerialNumber == deviceId)
      {
        ROS_INFO("Serial number matches device ID. Opening camera.");
        found = true;
        m_camera.Open(*it);
        break;
      }
    }

    // If the camera has not been found by its serial number,
    // or if the serial number is not specified, try opening the first camera found.
    if (!found)
    {
      //ROS_INFO("[%s] Opening first camera found...", ros::this_node::getName().c_str());
      m_camera.OpenFirstCamera();
      ROS_INFO("[%s] First camera found opened!", ros::this_node::getName().c_str());
    }

    m_cameraId = std::string("blaze_") + m_camera.GetCameraInfo().strSerialNumber;

      // Display info about the connected camera.
    ROS_INFO("[%s] * Infos about connected camera", ros::this_node::getName().c_str());
    ROS_INFO("[%s] *** ID:                %s", ros::this_node::getName().c_str(), m_camera.GetCameraInfo().strDeviceID.c_str());
    ROS_INFO("[%s] *** Model name:        %s", ros::this_node::getName().c_str(), m_camera.GetCameraInfo().strModelName.c_str());
    ROS_INFO("[%s] *** Display name:      %s", ros::this_node::getName().c_str(), m_camera.GetCameraInfo().strDisplayName.c_str());
    ROS_INFO("[%s] *** Serial number:     %s", ros::this_node::getName().c_str(), m_camera.GetCameraInfo().strSerialNumber.c_str());
    ROS_INFO("[%s] *** User defined name: %s", ros::this_node::getName().c_str(), m_camera.GetCameraInfo().strUserDefinedName.c_str());
    ROS_INFO("[%s] *** IP address:        %s", ros::this_node::getName().c_str(), m_camera.GetCameraInfo().strIpAddress.c_str());

    // Set up acquisition.
    GenApi::CEnumerationPtr ptrComponentSelector = m_camera.GetParameter("ComponentSelector");
    GenApi::CBooleanPtr ptrComponentEnable = m_camera.GetParameter("ComponentEnable");
    GenApi::CEnumerationPtr ptrPixelFormat = m_camera.GetParameter("PixelFormat");
    GenApi::CEnumerationPtr ptrScan3dCoordinateSelector = m_camera.GetParameter("Scan3dCoordinateSelector");
    GenApi::CFloatPtr ptrScan3dInvalidDataValue = m_camera.GetParameter("Scan3dInvalidDataValue");
    
    // Enable range data.
    ptrComponentSelector->FromString("Range");
    ptrComponentEnable->SetValue(true);
    ptrPixelFormat->FromString("Coord3D_ABC32f");   // 3D coordinates (point cloud)

    // Enable intensity image.
    ptrComponentSelector->FromString("Intensity");
    ptrPixelFormat->FromString("Mono16");
    ptrComponentEnable->SetValue(true);

    // Enable confidence map.
    ptrComponentSelector->FromString("Confidence");
    ptrPixelFormat->FromString("Confidence16");
    ptrComponentEnable->SetValue(true);
    
    m_invalidDataValue_old = ptrScan3dInvalidDataValue->GetValue();

    for (auto axis : { "CoordinateA", "CoordinateB", "CoordinateC" })
    {
        // Choose axis to configure.
        ptrScan3dCoordinateSelector->FromString(axis);
        // Set invalid data value for that axis.
        ptrScan3dInvalidDataValue->SetValue(s_invalidDataValue);
    }
  }
  catch(GenICam::GenericException& e)
  {
    ROS_ERROR("[%s] Error occurred when initializing camera", ros::this_node::getName().c_str());
    ROS_ERROR_STREAM("** Exception occurred: " << endl << e.GetDescription());
    exitCode = EXIT_FAILURE;
  }
  
  return exitCode;
}

void BlazeCamera::TerminateProcess()
{
  try
  {
    if ( m_camera.IsOpen() ) 
    {
        GenApi::CEnumerationPtr ptrScan3dCoordinateSelector = m_camera.GetParameter("Scan3dCoordinateSelector");
        GenApi::CFloatPtr ptrScan3dInvalidDataValue = m_camera.GetParameter("Scan3dInvalidDataValue");

        for (auto axis : {"CoordinateA", "CoordinateB", "CoordinateC"}) {
            // Choose axis to configure.
            ptrScan3dCoordinateSelector->FromString(axis);
            // Set invalid data value for that axis.
            ptrScan3dInvalidDataValue->SetValue(m_invalidDataValue_old);
        }
    }
  }
  catch(GenICam::GenericException& e)
  {
    ROS_ERROR_STREAM("Failed to reset properties to default: " << endl << e.GetDescription());
  }
  
  try
  {
    // Closing camera.
    if ( m_camera.IsOpen() )
    {
        m_camera.Close();
    }
  }
  catch(GenICam::GenericException& e)
  {
    ROS_ERROR_STREAM("Failed to properly close camera: " << endl << e.GetDescription());
  }  

  // Release the GenTL producer and all of its resources. 
  // Note: Don't call TerminateProducer() until the destructor of the CBlazeCamera
  // class has been called. The destructor may require resources which may not
  // be available anymore after TerminateProducer() has been called.
  if (CBlazeCamera::IsProducerInitialized())
      CBlazeCamera::TerminateProducer();  // Won't throw any exceptions.
}

void BlazeCamera::calculateRangeMap(const PartInfo& pointCloud, int minDepth, int maxDepth, uint16_t* pRangeMap)
{
    const int width = (int)pointCloud.width;
    const int height = (int)pointCloud.height;
    CBlazeCamera::Coord3D *pPoint = reinterpret_cast<CBlazeCamera::Coord3D*>(pointCloud.pData);

    const double scale = 65535.0 / (maxDepth - minDepth);

    for (int row = 0; row < height; ++row)
    {
        for (int col = 0; col < width; ++col, ++pPoint, ++pRangeMap)
        {
            if (isValid(pPoint))
            {
                // Calculate the radial distance.
                double distance = sqrt(pPoint->x * pPoint->x + pPoint->y * pPoint->y + pPoint->z * pPoint->z);
                // Clip to [minDepth..MaxDept].
                if (distance < minDepth)
                    distance = minDepth;
                else if (distance > maxDepth)
                    distance = maxDepth;
                *pRangeMap = (uint16_t) ( ( distance - minDepth ) * scale );
            }
            else
            {
                // No depth information available for this pixel. Zero it.
                *pRangeMap = 0;
            }
        }
    }
}

void BlazeCamera::calculateRangeMapColor(const PartInfo& pointCloud, int minDepth, int maxDepth, BGR* pRangeMap)
{
    const int width = (int) pointCloud.width;
    const int height = (int) pointCloud.height;
    CBlazeCamera::Coord3D *pPoint = reinterpret_cast<CBlazeCamera::Coord3D*>(pointCloud.pData);

    const double scale = 65535.0 / (maxDepth - minDepth);

    for (int row = 0; row < height; ++row)
    {
        for (int col = 0; col < width; ++col, ++pPoint, ++pRangeMap)
        {
            if (isValid(pPoint))
            {
                // Calculate the radial distance.
                double distance = sqrt(pPoint->x * pPoint->x + pPoint->y * pPoint->y + pPoint->z * pPoint->z);

                // Clip to [minDepth..MaxDept].
                if (distance < minDepth)
                    distance = minDepth;
                else if (distance > maxDepth)
                    distance = maxDepth;
                
                // Calculate the color.
                BGR bgr;
                const uint16_t g = (uint16_t)((distance - minDepth) * scale);
                const uint16_t val = g >> 6 & 0xff;
                const uint16_t sel = g >> 14;
                uint32_t res = val << 8 | 0xff;
                if (sel & 0x01)
                {
                    res = (~res) >> 8 & 0xffff;
                }
                if (sel & 0x02)
                {
                    res = res << 8;
                }
                bgr.r = res & 0xff;
                res = res >> 8;
                bgr.g = res & 0xff;
                res = res >> 8;
                bgr.b = res & 0xff;

                *pRangeMap = bgr;
            }
            else
            {
                // No depth information available for this pixel. Set it to black.
                BGR bgr;
                bgr.r = bgr.g = bgr.b = 0;
                *pRangeMap = bgr;
            }
        }
    }
}

PointCloudPtr BlazeCamera::convertGrabResultToPointCloud(const BufferParts& imgParts)
{
    // An organized point cloud is used, i.e., for each camera pixel there is an entry 
    // in the data structure indicating the 3D coordinates calculated from that pixel.
    // If the camera wasn't able to create depth information for a pixel, the x, y, and z coordinates 
    // are set to NaN. These NaNs will be retained in the PCL point cloud.

    // Allocate PCL point cloud.
    const size_t width = imgParts[0].width;
    const size_t height = imgParts[0].height;
    PointCloudPtr ptrPointCloud(new PointCloud_t);
    ptrPointCloud->width = width;
    ptrPointCloud->height = height;
    ptrPointCloud->points.resize(width * height);
    ptrPointCloud->is_dense = false; // Organized point cloud

    // Create a pointer to the 3D coordinates of the first point.
    // imgParts[0] always refers to the point cloud data.
    CBlazeCamera::Coord3D* pSrcPoint = (CBlazeCamera::Coord3D*) imgParts[0].pData;

    // Create a pointer to the intensity information stored in the second buffer part.
    uint16_t* pIntensity = (uint16_t*)imgParts[1].pData;

    // Set the points.
    for (size_t i = 0; i < height * width; ++i, ++pSrcPoint, ++pIntensity)
    {
        // Set the X/Y/Z cordinates.
        Point_t& dstPoint = ptrPointCloud->points[i];

        dstPoint.x = pSrcPoint->x * 0.001;
        dstPoint.y = pSrcPoint->y * 0.001;
        dstPoint.z = pSrcPoint->z * 0.001;

        // Use the intensity value of the pixel for coloring the point.
        dstPoint.r = dstPoint.g = dstPoint.b = (uint8_t)(*pIntensity >> 8);
    }

    return ptrPointCloud;
}

bool BlazeCamera::publish(const BufferParts& parts, ros::Time acquisitionTime)
{
  // Run some checks first.

  if (parts.size() != 3)
  {
    ROS_ERROR("[%s] Expected 3 parts, got %zu!", ros::this_node::getName().c_str(), parts.size());
    return EXIT_FAILURE;
  }

  if (parts[0].dataFormat != PFNC_Coord3D_ABC32f)
  {
    ROS_ERROR("[%s] Unexpected data format for the first image part. Coord3D_ABC32f is expected.", ros::this_node::getName().c_str());
    return EXIT_FAILURE;
  }

  if (parts[1].dataFormat != PFNC_Mono16)
  {
    ROS_ERROR("[%s] Unexpected data format for the second image part. Mono16 is expected.", ros::this_node::getName().c_str());
    return EXIT_FAILURE;
  }

  // Processing data.

  const int width = (int)parts[0].width;
  const int height = (int)parts[0].height;
  const int count = width * height;

  // Get min and max depth values.
  int minDepth = (int) GenApi::CIntegerPtr(m_camera.GetParameter("DepthMin"))->GetValue();
  int maxDepth = (int) GenApi::CIntegerPtr(m_camera.GetParameter("DepthMax"))->GetValue();

  // Point cloud
  if (m_cloudPub.getNumSubscribers() > 0)
  {
    PointCloudPtr ptrCloud = convertGrabResultToPointCloud(parts);
    ptrCloud->header.frame_id = g_frameId;
    ptrCloud->header.stamp = pcl_conversions::toPCL(acquisitionTime);
    ptrCloud->width = width;
    ptrCloud->height = height;
    ptrCloud->is_dense = false;
    ptrCloud->points.resize(count);
    // Publish
    m_cloudPub.publish(ptrCloud);

    sensor_msgs::CameraInfoPtr cloudInfoMsg(new sensor_msgs::CameraInfo(m_cloudInfoManager->getCameraInfo()));
    cloudInfoMsg->header.stamp    = acquisitionTime;
    cloudInfoMsg->header.frame_id = g_frameId;
    m_cloudCIPub.publish(cloudInfoMsg);
  }

  // Intensity
  if (m_intensityPub.getNumSubscribers() > 0)
  {
    cv::Mat intensity = cv::Mat(height, width, CV_16UC1, parts[1].pData);
    // Scale the intensity image since it often looks quite dark.
    double  max;
    cv::minMaxLoc(intensity, NULL, &max);
    intensity /= (max / std::numeric_limits<uint16_t>::max());
    // Publish
    cv_bridge::CvImage intensityCvImg;
    intensityCvImg.header.frame_id  = g_frameId;
    intensityCvImg.header.stamp     = acquisitionTime;
    intensityCvImg.encoding         = sensor_msgs::image_encodings::MONO16;
    intensityCvImg.image            = intensity;
    m_intensityPub.publish(intensityCvImg.toImageMsg());

    sensor_msgs::CameraInfoPtr intensityInfoMsg(new sensor_msgs::CameraInfo(m_intensityInfoManager->getCameraInfo()));
    intensityInfoMsg->header.stamp    = acquisitionTime;
    intensityInfoMsg->header.frame_id = g_frameId;
    m_intensityCIPub.publish(intensityInfoMsg);
  }

  // Range map
  if (m_rangeMapPub.getNumSubscribers() > 0)
  {
    uint16_t* pRangeData = new uint16_t[width * height];
    calculateRangeMap(parts[0], minDepth, maxDepth, pRangeData);
    cv::Mat rangeMap = cv::Mat(height, width, CV_16UC1, pRangeData);
    // Publish
    cv_bridge::CvImage rangeMapCvImg;
    rangeMapCvImg.header.frame_id  = g_frameId;
    rangeMapCvImg.header.stamp     = acquisitionTime;
    rangeMapCvImg.encoding         = sensor_msgs::image_encodings::MONO16;
    rangeMapCvImg.image            = rangeMap;
    m_rangeMapPub.publish(rangeMapCvImg.toImageMsg());

    sensor_msgs::CameraInfoPtr rangeMapInfoMsg(new sensor_msgs::CameraInfo(m_rangeMapInfoManager->getCameraInfo()));
    rangeMapInfoMsg->header.stamp    = acquisitionTime;
    rangeMapInfoMsg->header.frame_id = g_frameId;
    m_rangeMapCIPub.publish(rangeMapInfoMsg);

    delete[] pRangeData;
  }
  
  // Range map color
  if (m_rangeMapColorPub.getNumSubscribers() > 0)
  {
    BGR* pRangeDataColor = new BGR[width * height];
    calculateRangeMapColor(parts[0], minDepth, maxDepth, pRangeDataColor);
    cv::Mat3b rangeMapColor = cv::Mat(height, width, CV_8UC3, pRangeDataColor);
    // Publish
    cv_bridge::CvImage rangeMapColorCvImg;
    rangeMapColorCvImg.header.frame_id  = g_frameId;
    rangeMapColorCvImg.header.stamp     = acquisitionTime;
    rangeMapColorCvImg.encoding         = sensor_msgs::image_encodings::BGR8;
    rangeMapColorCvImg.image            = rangeMapColor;
    m_rangeMapColorPub.publish(rangeMapColorCvImg.toImageMsg());

    sensor_msgs::CameraInfoPtr rangeMapColorInfoMsg(new sensor_msgs::CameraInfo(m_rangeMapColorInfoManager->getCameraInfo()));
    rangeMapColorInfoMsg->header.stamp    = acquisitionTime;
    rangeMapColorInfoMsg->header.frame_id = g_frameId;
    m_rangeMapColorCIPub.publish(rangeMapColorInfoMsg);

    delete[] pRangeDataColor;
  }

  // Confidence map
  if (m_confidencePub.getNumSubscribers() > 0)
  {
    cv::Mat confidence = cv::Mat(height, width, CV_16UC1, parts[2].pData);
    // Publish
    cv_bridge::CvImage confidenceCvImg;
    confidenceCvImg.header.frame_id  = g_frameId;
    confidenceCvImg.header.stamp     = acquisitionTime;
    confidenceCvImg.encoding         = sensor_msgs::image_encodings::MONO16;
    confidenceCvImg.image            = confidence;
    m_confidencePub.publish(confidenceCvImg.toImageMsg());

    sensor_msgs::CameraInfoPtr confidenceInfoMsg(new sensor_msgs::CameraInfo(m_confidenceInfoManager->getCameraInfo()));
    confidenceInfoMsg->header.stamp    = acquisitionTime;
    confidenceInfoMsg->header.frame_id = g_frameId;
    m_confidenceCIPub.publish(confidenceInfoMsg);
  }

  return true;
}

int BlazeCamera::Run()
{
  int exitCode = EXIT_SUCCESS;

  try
  {
    const size_t nBuffers = 3;  // Number of buffers to be used for grabbing

    // Allocate the memory buffers and prepare image acquisition.
    m_camera.PrepareAcquisition(nBuffers);

    // Enqueue all buffers to be filled with image data.
    for (size_t i = 0; i < nBuffers; ++i)
    {
        m_camera.QueueBuffer(i);
    }

    // Start the acquisition engine.
    m_camera.StartAcquisition();
    m_camera.IssueAcquisitionStartCommand(); // The camera continuously sends data now.

    ROS_INFO("[%s] Acquisition started!", ros::this_node::getName().c_str());
    std::cout << std::endl;

    while (ros::ok())
    {
      ros::spinOnce();

      GrabResult grabResult;
      // Get next acquired frame frome the driver's output queue.
      m_camera.GetGrabResult(grabResult, 1000);

      if (grabResult.status == GrabResult::Timeout)
      {
          ROS_ERROR("[%s] Timeout occurred. Acquisition stopped.", ros::this_node::getName().c_str());
          // The timeout might be caused by a removal of the camera. Check if the camera is still connected.
          if (!m_camera.IsConnected())
          {
            ROS_WARN("[%s] Camera has been removed.", ros::this_node::getName().c_str());
            ROS_INFO("[%s] Stopping acquisition.", ros::this_node::getName().c_str());
            return exitCode;
          }
      }
      else
      {
        if (grabResult.status != GrabResult::Ok)
        {
          ROS_ERROR("[%s] Failed to grab image.", ros::this_node::getName().c_str());
        }
        else
        {
          BufferParts parts;
          m_camera.GetBufferParts(grabResult, parts);

          // Publish results if subscribers
          publish(parts, ros::Time::now());
        }

        m_camera.QueueBuffer(grabResult.hBuffer);
      }
    }
  }
  catch(GenICam::GenericException& e)
  {
    ROS_ERROR_STREAM("Exception occurred: " << endl << e.GetDescription());
    exitCode = EXIT_FAILURE;
  }

  return exitCode;
}

int main(int argc, char **argv)
{
  ROS_WARN("Node is initializing...");

  // Initialize node.
  ros::init(argc, argv, node_name.c_str());
  ros::NodeHandle n;
  // Parameter server
  ros::NodeHandle pn("~");

  // Camera info node handles
  ros::NodeHandle cloudNh("cloud");
  ros::NodeHandle intensityNh("intensity");
  ros::NodeHandle rangeMapNh("range_map");
  ros::NodeHandle rangeMapColorNh("range_map_color");
  ros::NodeHandle confidenceNh("confidence");

  if (!pn.getParam("device_id", g_deviceId))
  {
    ROS_WARN("~device_id is not set! Will try to connect to first camera found.");
    g_deviceId = "#1";
  }

  pn.param("frame_id", g_frameId, std::string("camera_optical_frame"));

  if (!pn.getParam("camera", g_namespace))
  {
    ROS_WARN("~camera is not set! Setting namespace to default value \'camera\'.");
    g_namespace = "camera";
  }

  // blaze camera instance
  BlazeCamera blazeCamera;

  // Initialize camera.
  int exitCode = blazeCamera.InitializeCamera(g_deviceId);
  if (exitCode == EXIT_FAILURE)
  {
    blazeCamera.TerminateProcess();
    std::cout << "[" << ros::this_node::getName().c_str() << "] Node is terminated with exit code = " << exitCode << std::endl;

    return exitCode;
  }

  // Topics
  blazeCamera.m_cloudPub = n.advertise<pcl::PointCloud<pcl::PointXYZRGB>>(g_cloudTopicName.c_str(), 10);
  blazeCamera.m_intensityPub = n.advertise<sensor_msgs::Image>(g_intensityTopicName.c_str(), 10);
  blazeCamera.m_rangeMapPub = n.advertise<sensor_msgs::Image>(g_rangeMapTopicName.c_str(), 10);
  blazeCamera.m_rangeMapColorPub = n.advertise<sensor_msgs::Image>(g_rangeMapColorTopicName.c_str(), 10);
  blazeCamera.m_confidencePub = n.advertise<sensor_msgs::Image>(g_confidenceTopicName.c_str(), 10);

  blazeCamera.m_cloudCIPub = n.advertise<sensor_msgs::CameraInfo>(g_cloudCITopicName.c_str(), 10);
  blazeCamera.m_intensityCIPub = n.advertise<sensor_msgs::CameraInfo>(g_intensityCITopicName.c_str(), 10);
  blazeCamera.m_rangeMapCIPub = n.advertise<sensor_msgs::CameraInfo>(g_rangeMapCITopicName.c_str(), 10);
  blazeCamera.m_rangeMapColorCIPub = n.advertise<sensor_msgs::CameraInfo>(g_rangeMapColorCITopicName.c_str(), 10);
  blazeCamera.m_confidenceCIPub = n.advertise<sensor_msgs::CameraInfo>(g_confidenceCITopicName.c_str(), 10);

  // Services
  ros::ServiceServer set_blaze_parameter_srv = n.advertiseService(g_setParameterSrvName.c_str(), &BlazeCamera::SetParameter, &blazeCamera);
  ros::ServiceServer get_blaze_parameter_srv = n.advertiseService(g_getParameterSrvName.c_str(), &BlazeCamera::GetParameter, &blazeCamera);
  ros::ServiceServer set_blaze_data_stream_parameter_srv = n.advertiseService(g_setDSParameterSrvName.c_str(), &BlazeCamera::GetDataStreamParameter, &blazeCamera);
  ros::ServiceServer get_blaze_data_stream_parameter_srv = n.advertiseService(g_getDSParameterSrvName.c_str(), &BlazeCamera::GetDataStreamParameter, &blazeCamera);
  ros::ServiceServer execute_command_srv = n.advertiseService(g_execCmdSrvName.c_str(), &BlazeCamera::ExecuteCommand, &blazeCamera);

  // Display some information.
  std::string cloudTopicName = g_namespace + "/" + g_cloudTopicName;
  std::string intensityTopicName = g_namespace + "/" + g_intensityTopicName;
  std::string rangeMapTopicName = g_namespace + "/" + g_rangeMapTopicName;
  std::string rangeMapColorTopicName = g_namespace + "/" + g_rangeMapColorTopicName;
  std::string confidenceTopicName = g_namespace + "/" + g_confidenceTopicName;

  std::string cloudCITopicName = cloudTopicName + "/camera_info";
  std::string intensityCITopicName = intensityTopicName + "/camera_info";
  std::string rangeMapCITopicName = rangeMapTopicName + "/camera_info";
  std::string rangeMapColorCITopicName = rangeMapColorTopicName + "/camera_info";
  std::string confidenceCITopicName = confidenceTopicName + "/camera_info";

  std::string setParameterSrvName = g_namespace + "/" + g_setParameterSrvName;
  std::string getParameterSrvName = g_namespace + "/" + g_getParameterSrvName;
  std::string setDSParameterSrvName = g_namespace + "/" + g_setDSParameterSrvName;
  std::string getDSParameterSrvName = g_namespace + "/" + g_getDSParameterSrvName;
  std::string execCmdSrvName = g_namespace + "/" + g_execCmdSrvName;

  ROS_INFO("[%s] Node started!", ros::this_node::getName().c_str());
  std::cout << std::endl;

  ROS_INFO("[%s] -> Advertising 3D point cloud through topic: %s", ros::this_node::getName().c_str(), cloudTopicName.c_str());
  ROS_INFO("[%s] -> Advertising 3D point cloud camera info through topic: %s", ros::this_node::getName().c_str(), cloudCITopicName.c_str());

  ROS_INFO("[%s] -> Advertising intensity image through topic: %s", ros::this_node::getName().c_str(), intensityTopicName.c_str());
  ROS_INFO("[%s] -> Advertising intensity image camera info through topic: %s", ros::this_node::getName().c_str(), intensityCITopicName.c_str());

  ROS_INFO("[%s] -> Advertising range map image through topic: %s", ros::this_node::getName().c_str(), rangeMapTopicName.c_str());
  ROS_INFO("[%s] -> Advertising range map image camera info through topic: %s", ros::this_node::getName().c_str(), rangeMapCITopicName.c_str());

  ROS_INFO("[%s] -> Advertising range map color image through topic: %s", ros::this_node::getName().c_str(), rangeMapColorTopicName.c_str());
  ROS_INFO("[%s] -> Advertising range map color image camera info through topic: %s", ros::this_node::getName().c_str(), rangeMapColorCITopicName.c_str());

  ROS_INFO("[%s] -> Advertising confidence image through topic: %s", ros::this_node::getName().c_str(), confidenceTopicName.c_str());
  ROS_INFO("[%s] -> Advertising confidence image camera info through topic: %s", ros::this_node::getName().c_str(), confidenceCITopicName.c_str());

  ROS_INFO("[%s] -> Setting parameter through service: %s", ros::this_node::getName().c_str(), setParameterSrvName.c_str());
  ROS_INFO("[%s] -> Getting parameter through service: %s", ros::this_node::getName().c_str(), getParameterSrvName.c_str());
  ROS_INFO("[%s] -> Setting data stream parameter through service: %s", ros::this_node::getName().c_str(), setDSParameterSrvName.c_str());
  ROS_INFO("[%s] -> Getting data stream parameter through service: %s", ros::this_node::getName().c_str(), getDSParameterSrvName.c_str());
  ROS_INFO("[%s] -> Executing command through service: %s", ros::this_node::getName().c_str(), execCmdSrvName.c_str());

  // Camera info manager
  blazeCamera.m_cloudInfoManager = boost::make_shared<camera_info_manager::CameraInfoManager>(cloudNh);
  blazeCamera.m_intensityInfoManager = boost::make_shared<camera_info_manager::CameraInfoManager>(intensityNh);
  blazeCamera.m_rangeMapInfoManager = boost::make_shared<camera_info_manager::CameraInfoManager>(rangeMapNh);
  blazeCamera.m_rangeMapColorInfoManager = boost::make_shared<camera_info_manager::CameraInfoManager>(rangeMapColorNh);
  blazeCamera.m_confidenceInfoManager = boost::make_shared<camera_info_manager::CameraInfoManager>(confidenceNh);

  // Camera info
  sensor_msgs::CameraInfoPtr ci = blazeCamera.GetCameraInfo();
  if (ci)
  {
      //std::cout << "Reading camera calibration from camera" << std::endl;
      blazeCamera.m_cloudInfoManager->setCameraInfo(*ci);
      blazeCamera.m_intensityInfoManager->setCameraInfo(*ci);
      blazeCamera.m_rangeMapInfoManager->setCameraInfo(*ci);
      blazeCamera.m_rangeMapColorInfoManager->setCameraInfo(*ci);
      blazeCamera.m_confidenceInfoManager->setCameraInfo(*ci);
  }

  // Start acquisition.
  std::cout << std::endl;
  ROS_WARN("[%s] Starting acquisition...", ros::this_node::getName().c_str());
  blazeCamera.Run();
  
  blazeCamera.TerminateProcess();
  std::cout << "[" << ros::this_node::getName().c_str() << "] Node is terminated with exit code = " << exitCode << std::endl;

  return exitCode;
}
